#if !defined(DECCW_H)
#define DECCW_H

/***********************************************************************
This header file provides an interface that maps proposed decimal
floating-point types to binary types, for use until a C compiler with
decimal support is available.  This should allow developing prototypes
of decimal code.

With gcc-4.2 and later, the --enable-decimal-float configure-time option
makes it possible to build a compiler with partial support for such
arithmetic (floating-point operations, but no library support yet, not
even I/O).  Apart from IBM S/390 (z-Series) mainframes, all systems that
offer decimal floating-point arithmetic are likely to do so in software,
and are expected to provide all three (32-, 64-, and 128-bit)
precisions.

To simplify building software that uses these new types, so that no
compile-time -DHAVE_xxx settings are needed, for now, we use explicit
tests for predefined symbols to select suitable values for the HAVE_xxx
symbols.

NB: to display predefined symbols with gcc-3.x and later, use

	gcc -dM -E - < /dev/null | sort -u > foo

    instead of

	touch empty.c
	gcc -v -E empty.c | sort -u > foo

***********************************************************************/

#if defined(__GNUC__)

/*
** TO DO: Update these comments and this block of conditionals as
** decimal-arithmetic support becomes available and operates (mostly)
** correctly in new gcc-4.x releases on various platforms.
**
** Supported systems:
**
**	gcc-4.3-20061118	GNU/Linux AMD64 (Red Hat)
**	gcc-4.3-20070511	GNU/Linux IA-32 (Fedora, Red Hat, Ubuntu) (unreliable)
**
**	gcc-4.4-20081017	GNU/Linux AMD64 (Red Hat)
**	gcc-4.4-20090123	Mac OS X PowerPC (__ppc__) and EM64T (with gcc source code patches)
**	gcc-4.4-20090123	FreeBSD 6 IA-32 (with gcc source code patches)
**	gcc-4.4-20090123	GNU/Linux IA-32 (Fedora, Red Hat, Ubuntu) (somewhat usable)
**
**	gcc-4.5-20090528	GNU/Linux AMD64 (Red Hat)
**	gcc-4.5-20090528	GNU/Linux IA-32 (Fedora and Ubuntu, copied to Red Hat)
**	gcc-4.5-20090528	GNU/Linux PowerPC (__PPC__)
**	gcc-4.5-20090528	Mac OS X PowerPC (__ppc__) and EM64T (no patches needed)
**
** gcc releases 4.3, 4.4, and 4.5 build on GNU/Linux IA-64, but the
** compiler crashes with segment violations in decimal code
** generation.
*/

#if (__GNUC__ >= 4) && defined(__DEC_EVAL_METHOD__) && \
    ( defined(__amd64) || defined(__amd64__) || \
      defined(__x86_64) || defined(__x86_64__) || \
      defined(__ppc__) || defined(__PPC__) || \
      defined(__i386__) ) && \
      !defined(HAVE_BROKEN_DECIMAL) && \
      !defined(__MMIX__) && \
      !defined(__PATHCC__) && \
      !defined(__UPC__)

#define HAVE_DECIMAL		1
#define HAVE_NATIVE_DECIMAL	1
#define HAVE__DECIMAL32		1
#define HAVE__DECIMAL64		1
#define HAVE__DECIMAL128	1

#undef  HAVE_DECIMAL_COMPLEX	/* not supported by gcc-4.2 */

#elif (__GNUC__ == 4) && (__GNUC_MINOR__ == 4) && defined(__ia64__) /* gcc-4.4 on IA-64 */

#define HAVE_DECIMAL		1
#define HAVE_NATIVE_DECIMAL	1
#define HAVE__DECIMAL32		1
#define HAVE__DECIMAL64		1
#define HAVE__DECIMAL128	1

#undef  HAVE_DECIMAL_COMPLEX

#endif /* (__GNUC__ >= 4) && defined(__DEC_EVAL_METHOD__) && ... */

#if defined(__MMIX__) || defined(__OPENCC__) || defined(__PATHCC__) || defined(__UPC__)
#define HAVE_BROKEN_DECIMAL
#endif

#if defined(HAVE_BROKEN_DECIMAL)

#undef HAVE_DECIMAL
#undef HAVE_NATIVE_DECIMAL
#undef HAVE__DECIMAL32
#undef HAVE__DECIMAL64
#undef HAVE__DECIMAL128
#undef HAVE__DECIMAL256

/*
** Avoid symbol conflict with compilers that know about the _Decimalxx
** type names, but have not yet implemented them properly, with the result
** that they cannot compile code where those types are visible.
*/
#define _Decimal32	_DECIMAL32
#define _Decimal64	_DECIMAL64
#define _Decimal128	_DECIMAL128
#define _Decimal256	_DECIMAL256

#endif /* defined(HAVE_BROKEN_DECIMAL) */

#endif /* defined(__GNUC__) */

/* First we have fallback typedefs for proposed decimal types */

#if !defined(HAVE__DECIMAL32)
typedef float _Decimal32;
#define	HAVE__DECIMAL32		1
#endif

#if !defined(HAVE__DECIMAL64)
typedef double _Decimal64;
#define HAVE__DECIMAL64		1
#endif

#if !defined(HAVE__DECIMAL128)

#if defined(HAVE_LONG_DOUBLE)
typedef long double _Decimal128;
#else
typedef _Decimal64 _Decimal128;
#endif

#define HAVE__DECIMAL128	1

#endif /* !defined(HAVE__DECIMAL128) */

#if !defined(HAVE__DECIMAL256)
typedef _Decimal128 _Decimal256;
#define HAVE__DECIMAL256	1
#endif /* !defined(HAVE__DECIMAL256) */

/* Next we have typedefs for bitsize-independent aliases */

#if !defined(HAVE_DECIMAL_FLOAT)
typedef _Decimal32 decimal_float;
#define HAVE_DECIMAL_FLOAT		1
#endif

#if !defined(HAVE_DECIMAL_DOUBLE)
typedef _Decimal64 decimal_double;
#define HAVE_DECIMAL_DOUBLE		1
#endif

#if !defined(HAVE_DECIMAL_LONG_DOUBLE)
typedef _Decimal128 decimal_long_double;
#define HAVE_DECIMAL_LONG_DOUBLE	1
#endif

#if !defined(HAVE_DECIMAL_LONG_LONG_DOUBLE)
typedef _Decimal256 decimal_long_long_double;
#define HAVE_DECIMAL_LONG_LONG_DOUBLE	1
#endif

/* Finally, we have extended types */

#if 0 && defined(__ia64) && defined(__hpux) && defined(_FPWIDETYPES)

#if defined(HAVE_LONG_DOUBLE)

#if !defined(HAVE_FLOAT80)
typedef long double __float80;
#endif

#if !defined(HAVE_FLOAT128)
typedef long double __float128;
#endif

#endif /* defined(HAVE_LONG_DOUBLE) */

#endif /* 0 && defined(__ia64) && defined(__hpux) && defined(_FPWIDETYPES) */

#endif /* !defined(DECCW_H) */
